from chalice import Chalice
from mongoengine import connect

from .chalicelib.resources import app as resources

DATABASE_URI = 'mongomock://localhost:27017/db'

app = Chalice(app_name='test_app')
app.register_blueprint(resources)

connect(host=DATABASE_URI)


@app.route('/')
def health_check():
    return dict(greeting="I'm testing app!!!")



from mongoengine import Document, StringField

from agave.models import BaseModel


class TestModel(BaseModel, Document):
    id = StringField()
    secret_field = StringField()
    __test__ = False
    _hidden = ['secret_field']


def test_hide_field():
    model = TestModel(id='12345', secret_field='secret')
    model_dict = model.to_dict()
    assert model_dict['secret_field'] == '********'
    assert model_dict['id'] == '12345'



from typing import Optional

from cuenca_validations.types import QueryParams
from pydantic import BaseModel


class AccountQuery(QueryParams):
    name: Optional[str] = None
    user_id: Optional[str] = None
    is_active: Optional[bool] = None


class TransactionQuery(QueryParams):
    user_id: Optional[str] = None


class AccountRequest(BaseModel):
    name: str


class AccountUpdateRequest(BaseModel):
    name: str



from agave.filters import generic_query

from ..models.transactions import Transaction as TransactionModel
from ..validators import TransactionQuery
from .base import app


@app.resource('/transactions')
class Transaction:
    model = TransactionModel
    query_validator = TransactionQuery
    get_query_filter = generic_query



from typing import ClassVar, Dict

from ..lib.mongoengine.model_helpers import mongo_to_dict


class BaseModel:
    _excluded: ClassVar = []
    _hidden: ClassVar = []

    def __init__(self, *args, **values):
        return super().__init__(*args, **values)

    def to_dict(self) -> Dict:
        private_fields = [f for f in dir(self) if f.startswith('_')]
        excluded = self._excluded + private_fields
        mongo_dict: dict = mongo_to_dict(self, excluded)

        for field in self._hidden:
            mongo_dict[field] = '********'
        return mongo_dict

    def __repr__(self) -> str:
        return str(self.to_dict())  # pragma: no cover



from functools import wraps
from typing import Callable

from chalice import Blueprint


class AuthedBlueprint(Blueprint):
    """
    This dummy class is an example of Authentication/Authorization blueprint.

    """

    def route(self, path: str, **kwargs):
        """
        Builds route decorator with custom authentication.
        It is only a function wrapper for `Blueprint._register_handler` methods

        For this example we do not validate any credentials but
        your authentication logic could be implemented here.

        :param path:
        :param kwargs:
        :return:
        """

        def decorator(user_handler: Callable):
            @wraps(user_handler)
            def authed_handler(*args, **kwargs):
                # your authentication logic goes here
                # before execute `user_handler` function.
                self.current_request.user_id = 'US123456789'
                return user_handler(*args, **kwargs)

            self._register_handler(
                'route',
                user_handler.__name__,
                authed_handler,
                authed_handler,
                dict(path=path, kwargs=kwargs),
            )

        return decorator

    def user_id_filter_required(self):
        """
        It overrides `RestApiBlueprint.user_id_filter_required()` method.

        This method is required to be implemented with your own business logic.
        You have to determine when `user_id` filter is required. For example:

        - `Account`s created by one user should not be queryable/retrievable
        by others users. In that case return `True`.

        - "Admin" users are allowed to query/retrieve any `Account` from any
        user. In that case return `False`.

        For testing purpose we return `False` as default behavior.
        But if we need to change it to `True` in tests we could monkey patch it
        when needed.

        :return:
        """
        return False



from typing import Any, Callable, Type


def copy_attributes(resource: Type[Any]):
    """
    Copy every attached property from resource methods definition to the
    real function handler.
    :param resource: Class representing the resource definition
    :return: wrapper function
    """

    def wrapper(func: Callable):
        try:
            original_func = getattr(resource, func.__name__)
        except AttributeError:
            return func

        for key, val in original_func.__dict__.items():
            setattr(func, key, val)

        return func

    return wrapper



from mongoengine import Document, FloatField, StringField

from agave.models import BaseModel
from agave.models.helpers import uuid_field


class Transaction(BaseModel, Document):
    id = StringField(primary_key=True, default=uuid_field('TR'))
    user_id = StringField(required=True)
    amount = FloatField(required=True)



import datetime as dt
from typing import Any

from blinker import NamedSignal
from mongoengine import signals


def handler(event: NamedSignal):
    """
    http://docs.mongoengine.org/guide/signals.html?highlight=update
    Signal decorator to allow use of callback functions as class
    decorators
    """

    def decorator(fn: Any):
        def apply(cls):
            event.connect(fn, sender=cls)
            return cls

        fn.apply = apply
        return fn

    return decorator


@handler(signals.pre_save)
def updated_at(_, document):
    document.updated_at = dt.datetime.utcnow()



import datetime as dt

from cuenca_validations.types import QueryParams

from agave.filters import generic_query


def test_generic_query_before():
    params = QueryParams(created_before=dt.datetime.utcnow().isoformat())
    query = generic_query(params)
    assert "created_at__lt" in repr(query)
    assert "user" not in repr(query)


def test_generic_query_after():
    params = QueryParams(created_after=dt.datetime.utcnow().isoformat())
    query = generic_query(params)
    assert "created_at__gt" in repr(query)
    assert "user" not in repr(query)



import datetime as dt

import pytest
from mongoengine import Document

from agave.lib.mongoengine.event_handlers import updated_at


@updated_at.apply
class TestModel(Document):
    __test__ = False


@pytest.mark.freeze_time('2020-10-10')
def test_attach_updated_at_field():
    model = TestModel()
    with pytest.raises(AttributeError):
        getattr(model, 'updated_at')

    model.save()
    assert type(model.updated_at) is dt.datetime
    assert model.updated_at == dt.datetime(2020, 10, 10)



from typing import Dict

from ..blueprints import AuthedRestApiBlueprint

app = AuthedRestApiBlueprint(__name__)


@app.get('/healthy_auth')
def health_auth_check() -> Dict:
    return dict(greeting="I'm authenticated and healthy !!!")



from typing import Optional, Type
from urllib.parse import urlencode

from chalice import Blueprint, NotFoundError, Response
from cuenca_validations.types import QueryParams
from mongoengine import DoesNotExist, Q
from pydantic import BaseModel, ValidationError

from .decorators import copy_attributes


class RestApiBlueprint(Blueprint):
    def get(self, path: str, **kwargs):
        return self.route(path, methods=['GET'], **kwargs)

    def post(self, path: str, **kwargs):
        return self.route(path, methods=['POST'], **kwargs)

    def patch(self, path: str, **kwargs):
        return self.route(path, methods=['PATCH'], **kwargs)

    def delete(self, path: str, **kwargs):
        return self.route(path, methods=['DELETE'], **kwargs)

    @property
    def current_user_id(self):
        return self.current_request.user_id

    def user_id_filter_required(self):
        """
        This method is required to be implemented with your own business logic.
        You are responsible of determining when `user_id` filter is required.
        """
        raise NotImplementedError(
            'this method should be override'
        )  # pragma: no cover

    def validate(self, validation_type: Type[BaseModel]):
        """This decorator validate the request body using a custom pydantyc model
        If validation fails return a BadRequest response with details

        @app.validate(MyPydanticModel)
        def my_method(request: MyPydanticModel):
            ...
        """

        def decorator(func):
            def wrapper(*args, **kwargs):
                try:
                    request = validation_type(**self.current_request.json_body)
                except ValidationError as e:
                    return Response(e.json(), status_code=400)
                return func(*args, request, **kwargs)

            return wrapper

        return decorator

    def resource(self, path: str):
        """Decorator to transform a class in Chalice REST endpoints

        @app.resource('/my_resource')
        class Items(Resource):
            model = MyMongoModel
            query_validator = MyPydanticModel

            def create(): ...
            def delete(id): ...
            def retrieve(id): ...
            def get_query_filter(): ...

        This implementation create the following endpoints

        POST /my_resource
        PATCH /my_resource
        DELETE /my_resource/id
        GET /my_resource/id
        GET /my_resource
        """

        def wrapper_resource_class(cls):
            """Wrapper for resource class
            :param cls: Resoucre class
            :return:
            """

            """ POST /resource
            Create a chalice endpoint using the method "create"
            If the method receive body params decorate it with @validate
            """
            if hasattr(cls, 'create'):
                route = self.post(path)
                route(cls.create)

            """ DELETE /resource/{id}
            Use "delete" method (if exists) to create the chalice endpoint
            """
            if hasattr(cls, 'delete'):
                route = self.delete(path + '/{id}')
                route(cls.delete)

            """ PATCH /resource/{id}
            Enable PATCH method if Resource.update method exist. It validates
            body data using `Resource.update_validator` but update logic is
            completely your responsibility.
            """
            if hasattr(cls, 'update'):
                route = self.patch(path + '/{id}')

                @copy_attributes(cls)
                def update(id: str):
                    params = self.current_request.json_body or dict()
                    try:
                        data = cls.update_validator(**params)
                        model = cls.model.objects.get(id=id)
                    except ValidationError as e:
                        return Response(e.json(), status_code=400)
                    except DoesNotExist:
                        raise NotFoundError('Not valid id')
                    else:
                        return cls.update(model, data)

                route(update)

            @self.get(path + '/{id}')
            @copy_attributes(cls)
            def retrieve(id: str):
                """GET /resource/{id}
                :param id: Object Id
                :return: Model object

                If exists "retrieve" method return the result of that, else
                use "id" param to retrieve the object of type "model" defined
                in the decorated class.

                The most of times this implementation is enough and is not
                necessary define a custom "retrieve" method
                """
                if hasattr(cls, 'retrieve'):
                    # at the moment, there are no resources with a custom
                    # retrieve method
                    return cls.retrieve(id)  # pragma: no cover
                try:
                    id_query = Q(id=id)
                    if self.user_id_filter_required():
                        id_query = id_query & Q(user_id=self.current_user_id)
                    data = cls.model.objects.get(id_query)
                except DoesNotExist:
                    raise NotFoundError('Not valid id')
                return data.to_dict()

            @self.get(path)
            @copy_attributes(cls)
            def query():
                """GET /resource
                Method for queries in resource. Use "query_validator" type
                defined in decorated class to validate the params.

                The "get_query_filter" method defined in decorated class
                should provide the way that the params are used to filter data

                If param "count" is True return the next response
                {
                    count:<count>
                }

                else the response is like this
                {
                    items = [{},{},...]
                    next_page = <url_for_next_items>
                }
                """
                params = self.current_request.query_params or dict()
                try:
                    query_params = cls.query_validator(**params)
                except ValidationError as e:
                    return Response(e.json(), status_code=400)
                # Set user_id request as query param
                if self.user_id_filter_required():
                    query_params.user_id = self.current_user_id
                filters = cls.get_query_filter(query_params)
                if query_params.count:
                    return _count(filters)
                return _all(query_params, filters)

            def _count(filters: Q):
                count = cls.model.objects.filter(filters).count()
                return dict(count=count)

            def _all(query: QueryParams, filters: Q):
                if query.limit:
                    limit = min(query.limit, query.page_size)
                    query.limit = max(0, query.limit - limit)  # type: ignore
                else:
                    limit = query.page_size
                items = (
                    cls.model.objects.order_by("-created_at")
                    .filter(filters)
                    .limit(limit)
                )
                item_dicts = [i.to_dict() for i in items]

                has_more: Optional[bool] = None
                if wants_more := query.limit is None or query.limit > 0:
                    # only perform this query if it's necessary
                    has_more = items.limit(limit + 1).count() > limit

                next_page_uri: Optional[str] = None
                if wants_more and has_more:
                    query.created_before = item_dicts[-1]['created_at']
                    path = self.current_request.context['resourcePath']
                    params = query.dict()
                    if self.user_id_filter_required():
                        params.pop('user_id')
                    next_page_uri = f'{path}?{urlencode(params)}'
                return dict(items=item_dicts, next_page_uri=next_page_uri)

            return cls

        return wrapper_resource_class



import functools
import json as jsonlib
from typing import Callable, Dict, Generator

FuncDecorator = Callable[..., Generator]


def auth_header(username: str, password: str) -> Dict:
    creds = username + password
    return {
        'Authorization': f'Basic {creds}',
    }


def accept_json(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(path, json=None, **kwargs):
        body = jsonlib.dumps(json) if json else None
        headers = {'Content-Type': 'application/json'}
        return func(path, body=body, headers=headers, **kwargs)

    return wrapper



from urllib.parse import urlencode

import pytest
from chalice.test import Client
from mock import MagicMock, patch

from examples.chalicelib.models import Account

USER_ID_FILTER_REQUIRED = (
    'examples.chalicelib.blueprints.authed.'
    'AuthedBlueprint.user_id_filter_required'
)


def test_create_resource(client: Client) -> None:
    data = dict(name='Doroteo Arango')
    resp = client.http.post('/accounts', json=data)
    model = Account.objects.get(id=resp.json_body['id'])
    assert resp.status_code == 201
    assert model.to_dict() == resp.json_body
    model.delete()


def test_create_resource_bad_request(client: Client) -> None:
    data = dict(invalid_field='some value')
    resp = client.http.post('/accounts', json=data)
    assert resp.status_code == 400


def test_retrieve_resource(client: Client, account: Account) -> None:
    resp = client.http.get(f'/accounts/{account.id}')
    assert resp.status_code == 200
    assert resp.json_body == account.to_dict()


@patch(USER_ID_FILTER_REQUIRED, MagicMock(return_value=True))
def test_retrieve_resource_user_id_filter_required(
    client: Client, other_account: Account
) -> None:
    resp = client.http.get(f'/accounts/{other_account.id}')
    assert resp.status_code == 404


def test_retrieve_resource_not_found(client: Client) -> None:
    resp = client.http.get('/accounts/unknown_id')
    assert resp.status_code == 404


def test_update_resource_with_invalid_params(client: Client) -> None:
    wrong_params = dict(wrong_param='wrong_value')
    response = client.http.patch(
        '/accounts/NOT_EXISTS',
        json=wrong_params,
    )
    assert response.status_code == 400


def test_update_resource_that_doesnt_exist(client: Client) -> None:
    resp = client.http.patch(
        '/accounts/5f9b4d0ff8d7255e3cc3c128',
        json=dict(name='Frida'),
    )
    assert resp.status_code == 404


def test_update_resource(client: Client, account: Account) -> None:
    resp = client.http.patch(
        f'/accounts/{account.id}',
        json=dict(name='Maria Felix'),
    )
    account.reload()
    assert resp.json_body['name'] == 'Maria Felix'
    assert account.name == 'Maria Felix'
    assert resp.status_code == 200


def test_delete_resource(client: Client, account: Account) -> None:
    resp = client.http.delete(f'/accounts/{account.id}')
    account.reload()
    assert resp.status_code == 200
    assert resp.json_body['deactivated_at'] is not None
    assert account.deactivated_at is not None


@pytest.mark.usefixtures('accounts')
def test_query_count_resource(client: Client) -> None:
    query_params = dict(count=1, name='Frida Kahlo')
    response = client.http.get(
        f'/accounts?{urlencode(query_params)}',
    )
    assert response.status_code == 200
    assert response.json_body['count'] == 1


@pytest.mark.usefixtures('accounts')
def test_query_all_with_limit(client: Client) -> None:
    limit = 2
    query_params = dict(limit=limit)
    response = client.http.get(f'/accounts?{urlencode(query_params)}')
    assert response.status_code == 200
    assert len(response.json_body['items']) == limit
    assert response.json_body['next_page_uri'] is None


@pytest.mark.usefixtures('accounts')
def test_query_all_resource(client: Client) -> None:
    query_params = dict(page_size=2)
    resp = client.http.get(f'/accounts?{urlencode(query_params)}')
    assert resp.status_code == 200
    assert len(resp.json_body['items']) == 2

    resp = client.http.get(resp.json_body['next_page_uri'])
    assert resp.status_code == 200
    assert len(resp.json_body['items']) == 2


@pytest.mark.usefixtures('accounts')
@patch(USER_ID_FILTER_REQUIRED, MagicMock(return_value=True))
def test_query_user_id_filter_required(client: Client) -> None:
    query_params = dict(page_size=2)
    resp = client.http.get(f'/accounts?{urlencode(query_params)}')
    assert resp.status_code == 200
    assert len(resp.json_body['items']) == 2
    assert all(
        item['user_id'] == 'US123456789' for item in resp.json_body['items']
    )

    resp = client.http.get(resp.json_body['next_page_uri'])
    assert resp.status_code == 200
    assert len(resp.json_body['items']) == 1
    assert all(
        item['user_id'] == 'US123456789' for item in resp.json_body['items']
    )


def test_query_resource_with_invalid_params(client: Client) -> None:
    wrong_params = dict(wrong_param='wrong_value')
    response = client.http.get(f'/accounts?{urlencode(wrong_params)}')
    assert response.status_code == 400


def test_cannot_create_resource(client: Client) -> None:
    response = client.http.post('/transactions', json=dict())
    assert response.status_code == 405


def test_cannot_update_resource(client: Client) -> None:
    response = client.http.post('/transactions', json=dict())
    assert response.status_code == 405


def test_cannot_delete_resource(client: Client) -> None:
    resp = client.http.delete('/transactions/TR1234')
    assert resp.status_code == 405
